(function () {

  chrome.storage.sync.get(null, function (shortcuts) {

    const isWF = url => /https:\/\/([a-z]+\.)?workflowy\.com(\/$|\/#)/.test(url);

    function runScript(str) {
      const s = document.createElement('script');
      s.textContent = str;
      (document.head || document.documentElement).appendChild(s);
      s.remove();
    }

    function runJS(shortcut) {
      let script;
      const jsType = shortcut[2];
      const run = shortcut[3];
      if (jsType === "rfw") {
        script = `
        (function () {
          const script = document.createElement('script');
          script.textContent = WF.getItemById('${run}').getNoteInPlainText();
          document.head.appendChild(script);
          script.remove();
        })();
        `;
      } else if (jsType === "run") {
        script = run
      } else {
        script = `
        (function () {
          let note = WF.getItemById('${run}').getNoteInPlainText();
          const obj = JSON.parse('${JSON.stringify(jsType)}');
          Object.keys(obj).forEach(key => {
            const rgx = new RegExp('(\\')(")?(_' + key + '_)(")?(\\')');
            note = note.replace(rgx, '$2' + obj[key] + '$4')
          });
          const script = document.createElement('script');
          script.textContent = note;
          document.head.appendChild(script);
          script.remove()
        })();`;
      }
      runScript(script)
    }

    /*
      MESSAGE/Shortcut
      [0] = name: "Name"
      [1] = ftoc: "domain", "path", "all", "js"
      [2] = update: true/false, "run", "rfw", {obj:setup}
      [3] = run: "url", "pID", "JavaScript"
    */

    function stayOrGo(shortcut) {
      if (isWF(shortcut[3]) && TAB_IS_WF) {
        location.href = new URL(shortcut[3]).hash
      } else if (shortcut[1] === "js" && shortcut[2] !== "run" && !TAB_IS_WF) {
        runScript("alert('This script only runs in a WorkFlowy tab.')")
      } else if (shortcut[1] === "js") {
        runJS(shortcut)
      } else {
        chrome.runtime.sendMessage(shortcut)
      }
    }

    // Returns decoded search or null. Requires valid shortcut as input
    function extractWfSearch(shortcut) {
      const run = shortcut[3];
      return isWF(run) && run.includes("?q=") ? decodeURIComponent(run.match(/(?<=\?q\=)(.+)/)[0]) : null;
    }

    function isLocalSearch(str) {
      const localMatch = str.match(/^( )(.*)/);
      if (localMatch === null || !TAB_IS_WF) return null;
      let localSearch = localMatch[2];
      const isLocalShortcut = shortcuts[localSearch];
      if (isLocalShortcut && extractWfSearch(isLocalShortcut)) localSearch = extractWfSearch(isLocalShortcut);
      return `WF.search(${JSON.stringify(localSearch)})`;
    }

    function globalSearch(str) {
      let search = `https://workflowy.com/#?q=${encodeURIComponent(str)}`;
      return stayOrGo([null, "domain", true, search]);
    }

    function shortcutMaker() {
      // getUserSetup returns obj for shortcuts[2]
      function getUserSetup(NAME) {
        let obj = {};
        const children = document.querySelectorAll(".children .task.noted");
        for (let child of children) {
          let header = `SETUP: ${NAME}\n`;
          let nMatch = child.getElementsByClassName("name")[0].textContent.match(/^(\w+)="(.*)"(_name_)?/);
          let note = child.getElementsByClassName("notes")[0].textContent.trimEnd();
          let p = prompt(header + note, nMatch[2].replace(/([@#])(\.)([a-zA-Z0-9])/g, '$1$3'));
          if (p === null) {
            obj = null;
            break;
          }
          if (nMatch[3]) {
            NAME = `${NAME}_${p}`;
            name = NAME;
          }
          // convert strings from prompt to integers and boolean
          if (/^\d+$/.test(p)) p = parseInt(p);
          if (/^(true|false)$/.test(p)) p = (p == "true");
          obj[nMatch[1]] = p;
        }
        return obj
      }

      const isInvalidShortcut = str => /(\s|'|")/.test(str) || /^(\?|wfs|wfo)$/.test(str) || str === "";

      function saveShortcut(sc) {
        const obj = {};
        obj[sc] = arrValue;
        if (confirm(`SAVE this shortcut?\n\nNAME = ${arrValue[0]}\nSHORTCUT = ${sc}\nRUN = ${runTxt}`)) {
          chrome.storage.sync.set(obj, function () {
            if (chrome.runtime.lastError) {
              alert(chrome.runtime.lastError.message)
            } else {
              chrome.runtime.sendMessage(["reload--options", "all", false, chrome.extension.getURL('options.html')]);
            }
          });
        }
      }
      function processMakerPrompt(shortcut) {
        if (shortcut === null) {
          return
        } else if (isInvalidShortcut(shortcut)) {
          if (confirm(`"${shortcut}" is not a valid shortcut.\n\nIt contains spaces, or is a reserved system shortcut.\n\nEnter a new shortcut.`)) {
            processMakerPrompt(prompt(promptText))
          } else {
            return
          }
        } else if (shortcuts[shortcut]) {
          if (confirm(`DUPLICATE SHORTCUT!\n\nSHORTCUT = ${shortcut}\nNAME = ${shortcuts[shortcut][0]}\n\nOK = REPLACE this shortcut with the current page.\nCANCEL = Define a NEW shortcut.`)) {
            saveShortcut(shortcut)
          } else {
            processMakerPrompt(prompt(promptText))
          }
        } else {
          saveShortcut(shortcut)
        }
      }
      function addSearch() {
        const q = location.href.match(/\?q=(.+)$/);
        return q ? ` - Search:${decodeURIComponent(q[1]).trim()}` : ""
      }

      let run = location.href;
      let ftoc = "domain";
      let update = true;
      let name;

      if (TAB_IS_WF) {
        const current = document.getElementsByClassName("project selected")[0];
        name = current.getElementsByClassName("name")[0].textContent.trim();
        const IS_BM = name.match(/( #js| #rfw| #setup)/);
        if (IS_BM) {
          name = name.replace(IS_BM[1], "").trimRight();
          ftoc = "js";
          run = current.getAttribute("projectid");
          if (IS_BM[1] === " #js") {
            update = "run";
            run = current.getElementsByClassName("notes")[0].textContent;
          } else if (IS_BM[1] === " #rfw") {
            update = "rfw";
          } else {
            update = getUserSetup(name);
            if (update === null) return
          }
        } else {
          // Append search info to WorkFlowy link if applicable
          name = name + addSearch();
        }
      } else {
        // External link
        name = document.title;
        update = false;
      }
      const arrValue = [name.trim(), ftoc, update, run];
      const runTxt = arrValue[1] === "js" ? "JavaScript" : arrValue[3];
      const promptText = `NAME:\n${arrValue[0]}\n\nRUN:\n${runTxt}\n\nSHORTCUT:`;
      processMakerPrompt(prompt(promptText));
    } // End Shortcut Maker

    function processPrompt(entry) {
      if (entry === null) {
        return;
      } else if (entry === '') {
        TAB_IS_WF ? runScript(goHomeJS) : chrome.runtime.sendMessage([null, "domain", false, "https://workflowy.com/"]);
      } else if (isLocalSearch(entry)) {
        runScript(isLocalSearch(entry))
      } else if (entry === '?') {
        console.table(shortcuts);
      } else if (entry === 'wfs') {
        shortcutMaker();
      } else if (entry === 'wfo') {
        chrome.runtime.sendMessage([null, "all", false, chrome.extension.getURL('options.html')]);
      } else if (shortcuts[entry]) {  // is Shortcut
        stayOrGo(shortcuts[entry]);
      } else if (entry.length <= 3 && !/^[#@]/.test(entry)) { // warnMeMax now hard coded at 3
        let warnMe = prompt(`\n"${entry}" is not a shortcut.\n\nPress OK to search or type a new entry.`, entry);
        if (warnMe === null) return;
        warnMe === entry ? globalSearch(warnMe) : processPrompt(warnMe);
      } else {
        globalSearch(entry);
      }
    }

    const goHomeJS = `
    (function(){
      const r=WF.rootItem();
      if(WF.currentSearchQuery())WF.clearSearch();
      WF.zoomTo(r);
      r.getVisibleChildren()[0].getElement().getElementsByClassName("content")[0].focus()})();
      `;
    const TAB_IS_WF = isWF(location.href);
    const selection = window.getSelection().toString();

    processPrompt(prompt('WFx', selection));

  });

})();