chrome.browserAction.onClicked.addListener(function (tab) {
  chrome.tabs.executeScript(tab.id, {
    file: "WFx.js"
  }, function () {
    // err happens when trying to execute the script on an "illegal" Chrome page. 
    let err = chrome.runtime.lastError;
    if (err) {
      console.log(err.message);
      findTabOrCreate("https://workflowy.com/", "domain", false);
    }
  });
});

/*
  MESSAGE/Shortcut
  [0] = name: "Name"
  [1] = ftoc: "domain", "path", "all", "js"
  [2] = update: boolean, "run", "rfw", {obj:setup}
  [3] = run: "url", "pID", "JavaScript"
*/

chrome.runtime.onMessage.addListener(function (message) {
  if (message[0] === "reload--options") {
    chrome.tabs.query({ url: message[3] }, function (tabs) {
      if (tabs.length > 0) chrome.tabs.reload(tabs[0].id)
    });
  } else {
    findTabOrCreate(message[3], message[1], message[2])
  }
});

const isWF = url => /https:\/\/([a-z]+\.)?workflowy\.com(\/$|\/#)/.test(url);

function findTabOrCreate(run, ftoc, update) {
  const getHash = url => new URL(url).hash;
  let tabMatch = ftocMatch(run, ftoc);
  chrome.tabs.query({}, function (tabs) {
    for (const tab of tabs) {
      if (isWF(tabMatch) && isWF(tab.url)) {
        const tabBase = new URL(tab.url).origin + "/";
        tabMatch = tabBase + getHash(tabMatch);
        run = tabBase + getHash(run);
      }
      if (tab.url.startsWith(tabMatch)) {
        let updateURL = update && run !== tab.url ? run : null;
        // Add hash to Shared WorkFlowy "home" url to prevent reload if update=true
        if (updateURL && /https:\/\/([a-z]+\.)?workflowy\.com\/s\//.test(updateURL) && !updateURL.includes('#')) updateURL = updateURL + '#';
        chrome.tabs.update(tab.id, { url: updateURL, active: true });
        chrome.windows.update(tab.windowId, { focused: true });
        return 
      }
    }
    chrome.tabs.create({ url: run, active: true });
  });
}

function ftocMatch(url, ftoc) {
  const IS_WF_OR_SHARE = /https:\/\/([a-z]+\.)?workflowy\.com(\/$|\/#|\/s\/)/.test(url);
  const loc = new URL(url);
  const path = loc.origin + loc.pathname;
  switch (ftoc) {
    case "domain":
      return IS_WF_OR_SHARE ? path : loc.origin;
    case "path":
      return IS_WF_OR_SHARE ? path + loc.hash.split("?")[0] : path;
    default:
      return url;
  }
}

chrome.contextMenus.create({
  title: "Find Tab Or Create",
  contexts: ["link", "selection"],
  id: "WFx"
});

chrome.contextMenus.onClicked.addListener(function (clickData) {
  const isValidUrl = (str) => {
    try {
      new URL(str);
      return true;
    } catch (_) {
      return false;
    }
  }
  let link = clickData.linkUrl;
  const selected = clickData.selectionText;
  if (selected && isValidUrl(selected) && !selected.startsWith("#")) link = selected;
  if (link) {
    // Fix for Google Calendar note link redirects
    const googleReDirect = link.match(/(url\?q=)(.+)(&sa=)/);
    if (googleReDirect) link = decodeURIComponent(googleReDirect[2]);
    return void findTabOrCreate(link, "domain", isWF(link))
  }
  if (selected) findTabOrCreate(`https://workflowy.com/#?q="${encodeURIComponent(selected)}"`, "domain", true);
});