(function () {

  function openTab(tabLink, tabName) {
    const tabs = document.getElementsByClassName("tabContent");
    for (const tab of tabs) {
      tab.style.display = "none";
    }
    const tabLinks = document.getElementsByClassName("tabLinks");
    for (const tablink of tabLinks) {
      tablink.classList.remove("active");
    }
    resetForms();
    document.getElementById(tabName).style.display = "block";
    document.getElementById(tabLink).classList.add("active");
    if (tabName === "shortcuts") document.getElementById("shortcutFilter").focus();
  }

  function resetForms() {
    document.getElementById("urlTab").textContent = "Add URL";
    document.getElementById("jsTab").textContent = "Add JavaScript";
    document.getElementById("urlName").value = "";
    document.getElementById("urlShortcut").value = "";
    document.getElementById("urlUrl").value = "";
    document.getElementById("ftoc").value = "domain";
    document.getElementById("update").checked = false;
    document.getElementById("deleteUrl").style.display = "none";
    document.getElementById("jsName").value = "";
    document.getElementById("jsShortcut").value = "";
    document.getElementById("jsText").value = "";
    document.getElementById("deleteJs").style.display = "none";
    document.getElementById("jsText").removeAttribute("readonly");
  }

  function openEditForm(key) {
    chrome.storage.sync.get([key], function (result) {
      oldShortcut = key;
      arr = result[key];
      const IS_JS = arr[1] === "js";
      document.getElementById(IS_JS ? "jsTab" : "urlTab").click();
      if (IS_JS) {
        document.getElementById("jsTab").textContent = "Edit JavaScript";
        document.getElementById("jsName").value = arr[0];
        document.getElementById("jsShortcut").value = key;
        document.getElementById("jsText").value = `${arr[2] !== "run" ? 'https://workflowy.com/#/' + arr[3].slice(-12) : arr[3]}${typeof arr[2] === "object" ? "\n\n" + JSON.stringify(arr[2], null, 2) : ""}`;
        if (arr[2] !== "run") document.getElementById("jsText").setAttribute("readonly", true);
        document.getElementById("deleteJs").style.display = "inline-block";
      } else {
        document.getElementById("urlTab").textContent = "Edit URL";
        document.getElementById("urlName").value = arr[0];
        document.getElementById("urlShortcut").value = key;
        document.getElementById("urlUrl").value = arr[3];
        document.getElementById("ftoc").value = arr[1];
        document.getElementById("update").checked = arr[2];
        document.getElementById("deleteUrl").style.display = "inline-block";
      }
    });
  }

  const isInvalidShortcut = str => /(\s|'|")/.test(str) || /^(\?|wfs|wfo)$/.test(str);
  
  function saveShortcut(IS_URL) {
    const IS_ADD = document.getElementById(IS_URL ? "deleteUrl" : "deleteJs").style.display === "none"; 
    const name = document.getElementById(IS_URL ? "urlName" : "jsName").value;
    const newShortcut = document.getElementById(IS_URL ? "urlShortcut" : "jsShortcut").value;
    const run = IS_URL ? document.getElementById("urlUrl").value : IS_ADD || arr[2] === "run" ? document.getElementById("jsText").value : arr[3];
    const ftoc = IS_URL ? document.getElementById("ftoc").value : "js";
    const update = IS_URL ? document.getElementById("update").checked : IS_ADD ? "run" : arr[2];
    let missing = [];
    if (name === "") missing.push("NAME");
    if (newShortcut === "") missing.push("SHORTCUT");
    if (run === "") missing.push(IS_URL ? "URL" : "JAVASCRIPT");
    if (missing.length > 0) return void alert(`Missing:\n${missing.join("\n")}`)
    if (isInvalidShortcut(newShortcut)) return void alert("INVALID Shortcut!\nTry again.");
    const SHORTCUT_EDITED = oldShortcut !== newShortcut;
    // Check for duplicate
    chrome.storage.sync.get([newShortcut], function (sc) {
      if (sc[newShortcut] && (IS_ADD || SHORTCUT_EDITED)) {
        return void alert("DUPLICATE Shortcut!\nTry again.")
      } else {
        // all info checks out... save (and delete)
        if (IS_ADD || !SHORTCUT_EDITED) {
          // just set shortcut data
          setShortcut(newShortcut, [name, ftoc, update, run])
        } else {
          // SHORTCUT_EDITED: set new data and delete old shortcut
          setShortcut(newShortcut, [name, ftoc, update, run], oldShortcut)
        }
      }
    });
  }

  function setShortcut(newSC, arrValue, oldSC) {
    const obj = {};
    obj[newSC] = arrValue;
    chrome.storage.sync.set(obj, function () {
      if (chrome.runtime.lastError) {
        alert(chrome.runtime.lastError.message)
      } else {
        if (!oldSC) {
          createTable();
          shortcutsTab.click();
        } else {
          chrome.storage.sync.remove([oldSC], function () {
            if (chrome.runtime.lastError) {
              alert(chrome.runtime.lastError.message)
            } else {
              createTable();
              shortcutsTab.click();
            }
          });
        }
      }
    });
  }

  function deleteShortcut(key) {
    if (confirm(`DELETE?\n${arr[0]}`)) {
      chrome.storage.sync.remove([key], function () {
        if (chrome.runtime.lastError) {
          alert(chrome.runtime.lastError.message)
        } else {
          createTable();
          shortcutsTab.click();
        }
      });
    }
  }

  function filterShortcuts() {
    let filter = document.getElementById("shortcutFilter").value.toLowerCase();
    const rows = document.getElementsByClassName("clickMe");
    for (let row of rows) {
      let td = row.getElementsByTagName("td");
      let combineRow = ` ${td[0].textContent} ${td[1].textContent} ${td[2].textContent}`;
      row.style.display = combineRow.toLowerCase().includes(filter) ? "" : "none";
    }
  }

  function importData() {
    let file = document.getElementById("file").files[0];
    let reader = new FileReader();
    reader.onload = function (e) {
      let obj = JSON.parse(e.target.result);
      for (const key in obj) {
        chrome.storage.sync.get([key], function (result) {
          if (result[key] === undefined) {
            chrome.storage.sync.set({ [key]: obj[key] }, function () {
              if (chrome.runtime.lastError) {
                console.log(chrome.runtime.lastError.message);
                error = true;
              }
            });
          }
        });
      }
    };
    reader.readAsText(file);
    setTimeout(() => {
      location.reload();
    }, 750);
  }

  function exportData() {
    chrome.storage.sync.get(null, function (items) {
      const date = new Date().toLocaleDateString("sv");
      const keyLength = Object.keys(items).length;
      const textData = JSON.stringify(items);
      const url = `data:text/json;charset=utf-8,${encodeURIComponent(textData)}`;
      chrome.downloads.download({
        url: url,
        filename: `WFx_export_${date}_${keyLength}.txt`
      });
    });
  }

  function createTable() {
    const eHTML = str => str.replace(/&/g, "&amp;").replace(/</g, "&lt;").replace(/>/g, "&gt;");
    const isWF = url => /https:\/\/([a-z]+\.)?workflowy\.com(\/$|\/#)/.test(url);
    const tbody = document.getElementById("tbody");
    tbody.innerHTML = "";
    chrome.storage.sync.get(null, function (shortcuts) {
      for (const shortcut in shortcuts) {
        let tArr, name, type;
        tArr = shortcuts[shortcut];
        name = tArr[0];
        type = tArr[1] === "js" ? "JS" : isWF(tArr[3]) ? "WF" : "Web";
        tbody.innerHTML += `<tr class="clickMe"><td>${eHTML(shortcut)}</td><td>${eHTML(name)}</td><td>${type}</tr>`;
      }
      // MAKE TABLE ROWS CLICKABLE 
      const trs = document.getElementsByClassName("clickMe");
      for (const tr of trs) {
        tr.onclick = function () { openEditForm(this.getElementsByTagName("td")[0].textContent) };
      }
      document.getElementById("shortcutFilter").value = "";
    });
  }

  let oldShortcut, arr; // set global

  window.addEventListener('DOMContentLoaded', () => {
    // SHORTCUT TAB 
    createTable();
    const shortcutsTab = document.getElementById("shortcutsTab");
    shortcutsTab.onclick = () => openTab('shortcutsTab', 'shortcuts');
    document.getElementById("shortcutFilter").onkeyup = () => filterShortcuts();

    // ADD URL TAB
    document.getElementById("urlTab").onclick = () => openTab('urlTab', 'addURL');
    document.getElementById("saveUrl").onclick = () => saveShortcut(true);
    document.getElementById("cancelUrl").onclick = () => shortcutsTab.click();
    document.getElementById("deleteUrl").onclick = () => deleteShortcut(oldShortcut);

    //  ADD JS TAB
    document.getElementById("jsTab").onclick = () => openTab('jsTab', 'addJS');
    document.getElementById("saveJs").onclick = () => saveShortcut(false);
    document.getElementById("cancelJs").onclick = () => shortcutsTab.click();
    document.getElementById("deleteJs").onclick = () => deleteShortcut(oldShortcut);

    // MISC SETTINGS TAB
    document.getElementById("miscTab").onclick = () => openTab('miscTab', 'misc');
    document.getElementById('setKBS').onclick = () => chrome.tabs.update({ url: 'chrome://extensions/shortcuts' });
    document.getElementById("importButton").onclick = () => importData();
    document.getElementById("exportButton").onclick = () => exportData();

    // HELP TAB
    document.getElementById("helpTab").onclick = () => openTab('helpTab', 'help');

    shortcutsTab.click();
    document.getElementById("shortcutFilter").focus();

    // ESC: FOCUS AND CLEAR SHORTCUT FILTER OR GO TO SHORTCUTS/CANCEL
    document.addEventListener("keydown", function (event) {
      if (!event.altKey && !event.ctrlKey && !event.shiftKey && !event.metaKey && event.key === "Escape") {
        if (document.getElementById("shortcutsTab").className.includes("active")) {
          const shortcutFilter = document.getElementById("shortcutFilter");
          if (shortcutFilter === document.activeElement) {
            shortcutFilter.value = "";
          } else {
            shortcutFilter.focus();
          }
        } else {
          shortcutsTab.click();
        }
      }
    });
  });
})();
